"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MangoClient = exports.MarginAccount = exports.MangoGroup = void 0;
const web3_js_1 = require("@solana/web3.js");
const layout_1 = require("./layout");
const bn_js_1 = __importDefault(require("bn.js"));
const utils_1 = require("./utils");
const serum_1 = require("@project-serum/serum");
const token_instructions_1 = require("@project-serum/serum/lib/token-instructions");
const instruction_1 = require("./instruction");
const schema_1 = require("./schema");
const spl_token_1 = require("@solana/spl-token");
class MangoGroup {
    constructor(publicKey, decoded, nativeSrm) {
        this.publicKey = publicKey;
        Object.assign(this, decoded);
        if (nativeSrm) {
            this.nativeSrm = nativeSrm;
        }
        else {
            this.nativeSrm = null;
        }
    }
    async getPrices(connection) {
        const aggs = await Promise.all(this.oracles.map((pk) => (schema_1.Aggregator.loadWithConnection(pk, connection))));
        return aggs.map((agg) => (agg.answer.median.toNumber() / Math.pow(10, agg.config.decimals))).concat(1.0);
    }
    getMarketIndex(spotMarket) {
        for (let i = 0; i < this.spotMarkets.length; i++) {
            if (this.spotMarkets[i].equals(spotMarket.publicKey)) {
                return i;
            }
        }
        throw new Error("This Market does not belong to this MangoGroup");
    }
    getTokenIndex(token) {
        for (let i = 0; i < this.tokens.length; i++) {
            if (this.tokens[i].equals(token)) {
                return i;
            }
        }
        throw new Error("This token does not belong in this MangoGroup");
    }
    getBorrowRate(tokenIndex) {
        const totalBorrows = this.getUiTotalBorrow(tokenIndex);
        const totalDeposits = this.getUiTotalDeposit(tokenIndex);
        if (totalDeposits === 0 && totalBorrows === 0) {
            return 0;
        }
        if (totalDeposits <= totalBorrows) {
            return layout_1.MAX_RATE;
        }
        const utilization = totalBorrows / totalDeposits;
        if (utilization > layout_1.OPTIMAL_UTIL) {
            const extraUtil = utilization - layout_1.OPTIMAL_UTIL;
            const slope = (layout_1.MAX_RATE - layout_1.OPTIMAL_RATE) / (1 - layout_1.OPTIMAL_UTIL);
            return layout_1.OPTIMAL_RATE + slope * extraUtil;
        }
        else {
            const slope = layout_1.OPTIMAL_RATE / layout_1.OPTIMAL_UTIL;
            return slope * utilization;
        }
    }
    getDepositRate(tokenIndex) {
        const borrowRate = this.getBorrowRate(tokenIndex);
        const totalBorrows = this.getUiTotalBorrow(tokenIndex);
        const totalDeposits = this.getUiTotalDeposit(tokenIndex);
        if (totalDeposits === 0 && totalBorrows === 0) {
            return 0;
        }
        else if (totalDeposits === 0) {
            return layout_1.MAX_RATE;
        }
        const utilization = totalBorrows / totalDeposits;
        return utilization * borrowRate;
    }
    getUiTotalDeposit(tokenIndex) {
        return utils_1.nativeToUi(this.totalDeposits[tokenIndex] * this.indexes[tokenIndex].deposit, this.mintDecimals[tokenIndex]);
    }
    getUiTotalBorrow(tokenIndex) {
        return utils_1.nativeToUi(this.totalBorrows[tokenIndex] * this.indexes[tokenIndex].borrow, this.mintDecimals[tokenIndex]);
    }
}
exports.MangoGroup = MangoGroup;
class MarginAccount {
    // TODO keep updated with websocket
    constructor(publicKey, decoded) {
        this.publicKey = publicKey;
        this.createTime = utils_1.getUnixTs();
        this.openOrdersAccounts = new Array(layout_1.NUM_MARKETS).fill(undefined);
        Object.assign(this, decoded);
    }
    getNativeDeposit(mangoGroup, tokenIndex) {
        return Math.round(mangoGroup.indexes[tokenIndex].deposit * this.deposits[tokenIndex]);
    }
    getNativeBorrow(mangoGroup, tokenIndex) {
        return Math.round(mangoGroup.indexes[tokenIndex].borrow * this.borrows[tokenIndex]);
    }
    getUiDeposit(mangoGroup, tokenIndex) {
        return utils_1.nativeToUi(this.getNativeDeposit(mangoGroup, tokenIndex), mangoGroup.mintDecimals[tokenIndex]);
    }
    getUiBorrow(mangoGroup, tokenIndex) {
        return utils_1.nativeToUi(this.getNativeBorrow(mangoGroup, tokenIndex), mangoGroup.mintDecimals[tokenIndex]);
    }
    getUiSrmBalance() {
        return utils_1.nativeToUi(this.srmBalance, token_instructions_1.SRM_DECIMALS);
    }
    async loadOpenOrders(connection, dexProgramId) {
        const promises = [];
        for (let i = 0; i < this.openOrders.length; i++) {
            if (this.openOrders[i].equals(utils_1.zeroKey)) {
                promises.push(utils_1.promiseUndef());
            }
            else {
                promises.push(serum_1.OpenOrders.load(connection, this.openOrders[i], dexProgramId));
            }
        }
        this.openOrdersAccounts = await Promise.all(promises);
        return this.openOrdersAccounts;
    }
    toPrettyString(mangoGroup, prices) {
        const lines = [
            `MarginAccount: ${this.publicKey.toBase58()}`,
            `${"Asset".padEnd(5)} ${"Deposits".padEnd(10)} ${"Borrows".padEnd(10)}`,
        ];
        const tokenToDecimals = {
            "BTC": 4,
            "ETH": 3,
            "USDC": 2
        };
        const tokenNames = ["BTC", "ETH", "USDC"]; // TODO pull this from somewhere
        for (let i = 0; i < mangoGroup.tokens.length; i++) {
            const decimals = tokenToDecimals[tokenNames[i]];
            const depositStr = this.getUiDeposit(mangoGroup, i).toFixed(decimals).toString().padEnd(10);
            const borrowStr = this.getUiBorrow(mangoGroup, i).toFixed(decimals).toString().padEnd(10);
            lines.push(`${tokenNames[i].padEnd(5)} ${depositStr} ${borrowStr}`);
        }
        lines.push(`Coll. Ratio: ${this.getCollateralRatio(mangoGroup, prices).toFixed(4)}`);
        lines.push(`Value: ${this.computeValue(mangoGroup, prices).toFixed(2)}`);
        return lines.join('\n');
    }
    computeValue(mangoGroup, prices) {
        let value = 0;
        for (let i = 0; i < this.deposits.length; i++) {
            value += (this.getUiDeposit(mangoGroup, i) - this.getUiBorrow(mangoGroup, i)) * prices[i];
        }
        for (let i = 0; i < this.openOrdersAccounts.length; i++) {
            const oos = this.openOrdersAccounts[i];
            if (oos != undefined) {
                value += utils_1.nativeToUi(oos.baseTokenTotal.toNumber(), mangoGroup.mintDecimals[i]) * prices[i];
                value += utils_1.nativeToUi(oos.quoteTokenTotal.toNumber(), mangoGroup.mintDecimals[layout_1.NUM_TOKENS - 1]);
            }
        }
        return value;
    }
    async getValue(connection, mangoGroup) {
        const prices = await mangoGroup.getPrices(connection);
        return this.computeValue(mangoGroup, prices);
    }
    getDeposits(mangoGroup) {
        const deposits = new Array(layout_1.NUM_TOKENS);
        for (let i = 0; i < layout_1.NUM_TOKENS; i++) {
            deposits[i] = this.getUiDeposit(mangoGroup, i);
        }
        return deposits;
    }
    getAssets(mangoGroup) {
        const assets = new Array(layout_1.NUM_TOKENS);
        for (let i = 0; i < layout_1.NUM_TOKENS; i++) {
            assets[i] = this.getUiDeposit(mangoGroup, i);
        }
        for (let i = 0; i < layout_1.NUM_MARKETS; i++) {
            const openOrdersAccount = this.openOrdersAccounts[i];
            if (openOrdersAccount == undefined) {
                continue;
            }
            assets[i] += utils_1.nativeToUi(openOrdersAccount.baseTokenTotal.toNumber(), mangoGroup.mintDecimals[i]);
            assets[layout_1.NUM_TOKENS - 1] += utils_1.nativeToUi(openOrdersAccount.quoteTokenTotal.toNumber(), mangoGroup.mintDecimals[layout_1.NUM_TOKENS - 1]);
        }
        return assets;
    }
    getLiabs(mangoGroup) {
        const liabs = new Array(layout_1.NUM_TOKENS);
        for (let i = 0; i < layout_1.NUM_TOKENS; i++) {
            liabs[i] = this.getUiBorrow(mangoGroup, i);
        }
        return liabs;
    }
    getAssetsVal(mangoGroup, prices) {
        let assetsVal = 0;
        for (let i = 0; i < layout_1.NUM_TOKENS; i++) {
            assetsVal += this.getUiDeposit(mangoGroup, i) * prices[i];
        }
        for (let i = 0; i < layout_1.NUM_MARKETS; i++) {
            const openOrdersAccount = this.openOrdersAccounts[i];
            if (openOrdersAccount == undefined) {
                continue;
            }
            assetsVal += utils_1.nativeToUi(openOrdersAccount.baseTokenTotal.toNumber(), mangoGroup.mintDecimals[i]) * prices[i];
            assetsVal += utils_1.nativeToUi(openOrdersAccount.quoteTokenTotal.toNumber(), mangoGroup.mintDecimals[layout_1.NUM_TOKENS - 1]);
        }
        return assetsVal;
    }
    getLiabsVal(mangoGroup, prices) {
        let liabsVal = 0;
        for (let i = 0; i < layout_1.NUM_TOKENS; i++) {
            liabsVal += this.getUiBorrow(mangoGroup, i) * prices[i];
        }
        return liabsVal;
    }
    getCollateralRatio(mangoGroup, prices) {
        const assetsVal = this.getAssetsVal(mangoGroup, prices);
        const liabsVal = this.getLiabsVal(mangoGroup, prices);
        return assetsVal / liabsVal;
    }
    async cancelAllOrdersByMarket(connection, client, programId, mangoGroup, market, bids, asks, owner) {
        const marketIndex = mangoGroup.getMarketIndex(market);
        const openOrdersAccount = this.openOrdersAccounts[marketIndex];
        if (openOrdersAccount == undefined) { // no open orders for this market
            return [];
        }
        const orders = market.filterForOpenOrders(bids, asks, [openOrdersAccount]);
        return await Promise.all(orders.map((order) => (client.cancelOrder(connection, programId, mangoGroup, this, owner, market, order))));
    }
}
exports.MarginAccount = MarginAccount;
class MangoClient {
    async sendTransaction2(connection, transaction, payer, additionalSigners) {
        transaction.recentBlockhash = (await connection.getRecentBlockhash('singleGossip')).blockhash;
        transaction.setSigners(payer.publicKey, ...additionalSigners.map(a => a.publicKey));
        const signers = [payer].concat(additionalSigners);
        transaction.sign(...signers);
        const rawTransaction = transaction.serialize();
        return await web3_js_1.sendAndConfirmRawTransaction(connection, rawTransaction, { skipPreflight: true });
    }
    async sendTransaction(connection, transaction, payer, additionalSigners) {
        // TODO test on mainnet
        transaction.recentBlockhash = (await connection.getRecentBlockhash('singleGossip')).blockhash;
        transaction.setSigners(payer.publicKey, ...additionalSigners.map(a => a.publicKey));
        const signers = [payer].concat(additionalSigners);
        transaction.sign(...signers);
        const rawTransaction = transaction.serialize();
        return await web3_js_1.sendAndConfirmRawTransaction(connection, rawTransaction, { skipPreflight: true });
    }
    async initMangoGroup(connection, programId, payer) {
        throw new Error("Not Implemented");
    }
    async initMarginAccount(connection, programId, mangoGroup, owner) {
        // Create a Solana account for the MarginAccount and allocate space
        const accInstr = await utils_1.createAccountInstruction(connection, owner.publicKey, layout_1.MarginAccountLayout.span, programId);
        // Specify the accounts this instruction takes in (see program/src/instruction.rs)
        const keys = [
            { isSigner: false, isWritable: false, pubkey: mangoGroup.publicKey },
            { isSigner: false, isWritable: true, pubkey: accInstr.account.publicKey },
            { isSigner: true, isWritable: false, pubkey: owner.publicKey },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_RENT_PUBKEY }
        ];
        // Encode and create instruction for actual initMarginAccount instruction
        const data = layout_1.encodeMangoInstruction({ InitMarginAccount: {} });
        const initMarginAccountInstruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        // Add all instructions to one atomic transaction
        const transaction = new web3_js_1.Transaction();
        transaction.add(accInstr.instruction);
        transaction.add(initMarginAccountInstruction);
        // Specify signers in addition to the wallet
        const additionalSigners = [
            accInstr.account,
        ];
        // sign, send and confirm transaction
        await this.sendTransaction(connection, transaction, owner, additionalSigners);
        return accInstr.account.publicKey;
    }
    async deposit(connection, programId, mangoGroup, marginAccount, owner, token, tokenAcc, quantity) {
        const tokenIndex = mangoGroup.getTokenIndex(token);
        const nativeQuantity = utils_1.uiToNative(quantity, mangoGroup.mintDecimals[tokenIndex]);
        const keys = [
            { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
            { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
            { isSigner: true, isWritable: false, pubkey: owner.publicKey },
            { isSigner: false, isWritable: true, pubkey: tokenAcc },
            { isSigner: false, isWritable: true, pubkey: mangoGroup.vaults[tokenIndex] },
            { isSigner: false, isWritable: false, pubkey: spl_token_1.TOKEN_PROGRAM_ID },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY }
        ];
        const data = layout_1.encodeMangoInstruction({ Deposit: { quantity: nativeQuantity } });
        const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        const additionalSigners = [];
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    async withdraw(connection, programId, mangoGroup, marginAccount, owner, token, tokenAcc, quantity) {
        const tokenIndex = mangoGroup.getTokenIndex(token);
        const nativeQuantity = utils_1.uiToNative(quantity, mangoGroup.mintDecimals[tokenIndex]);
        const keys = [
            { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
            { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
            { isSigner: true, isWritable: false, pubkey: owner.publicKey },
            { isSigner: false, isWritable: true, pubkey: tokenAcc },
            { isSigner: false, isWritable: true, pubkey: mangoGroup.vaults[tokenIndex] },
            { isSigner: false, isWritable: false, pubkey: mangoGroup.signerKey },
            { isSigner: false, isWritable: false, pubkey: spl_token_1.TOKEN_PROGRAM_ID },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY },
            ...marginAccount.openOrders.map((pubkey) => ({ isSigner: false, isWritable: false, pubkey })),
            ...mangoGroup.oracles.map((pubkey) => ({ isSigner: false, isWritable: false, pubkey }))
        ];
        const data = layout_1.encodeMangoInstruction({ Withdraw: { quantity: nativeQuantity } });
        const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        const additionalSigners = [];
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    async borrow(connection, programId, mangoGroup, marginAccount, owner, token, quantity) {
        const tokenIndex = mangoGroup.getTokenIndex(token);
        const nativeQuantity = utils_1.uiToNative(quantity, mangoGroup.mintDecimals[tokenIndex]);
        const keys = [
            { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
            { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
            { isSigner: true, isWritable: false, pubkey: owner.publicKey },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY },
            ...marginAccount.openOrders.map((pubkey) => ({ isSigner: false, isWritable: false, pubkey })),
            ...mangoGroup.oracles.map((pubkey) => ({ isSigner: false, isWritable: false, pubkey })),
        ];
        const data = layout_1.encodeMangoInstruction({ Borrow: { tokenIndex: new bn_js_1.default(tokenIndex), quantity: nativeQuantity } });
        const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        const additionalSigners = [];
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    async settleBorrow(connection, programId, mangoGroup, marginAccount, owner, token, quantity) {
        const tokenIndex = mangoGroup.getTokenIndex(token);
        const nativeQuantity = utils_1.uiToNative(quantity, mangoGroup.mintDecimals[tokenIndex]);
        const keys = [
            { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
            { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
            { isSigner: true, isWritable: false, pubkey: owner.publicKey },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY }
        ];
        const data = layout_1.encodeMangoInstruction({ SettleBorrow: { tokenIndex: new bn_js_1.default(tokenIndex), quantity: nativeQuantity } });
        const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        const additionalSigners = [];
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    /**
     * Call SettleFunds on each market, then call SettleBorrow for each token in one transaction
     * @param connection
     * @param programId
     * @param mangoGroup
     * @param marginAccount
     * @param markets
     * @param owner
     */
    async settleAll(connection, programId, mangoGroup, marginAccount, markets, owner) {
        const transaction = new web3_js_1.Transaction();
        const assetGains = new Array(layout_1.NUM_TOKENS).fill(0);
        for (let i = 0; i < layout_1.NUM_MARKETS; i++) {
            const openOrdersAccount = marginAccount.openOrdersAccounts[i];
            if (openOrdersAccount === undefined) {
                continue;
            }
            else if (openOrdersAccount.quoteTokenFree.toNumber() === 0 && openOrdersAccount.baseTokenFree.toNumber() === 0) {
                continue;
            }
            assetGains[i] += openOrdersAccount.baseTokenFree.toNumber();
            assetGains[layout_1.NUM_TOKENS - 1] += openOrdersAccount.quoteTokenFree.toNumber();
            const spotMarket = markets[i];
            const dexSigner = await web3_js_1.PublicKey.createProgramAddress([
                spotMarket.publicKey.toBuffer(),
                spotMarket['_decoded'].vaultSignerNonce.toArrayLike(Buffer, 'le', 8)
            ], spotMarket.programId);
            const keys = [
                { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
                { isSigner: true, isWritable: false, pubkey: owner.publicKey },
                { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
                { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY },
                { isSigner: false, isWritable: false, pubkey: spotMarket.programId },
                { isSigner: false, isWritable: true, pubkey: spotMarket.publicKey },
                { isSigner: false, isWritable: true, pubkey: marginAccount.openOrders[i] },
                { isSigner: false, isWritable: false, pubkey: mangoGroup.signerKey },
                { isSigner: false, isWritable: true, pubkey: spotMarket['_decoded'].baseVault },
                { isSigner: false, isWritable: true, pubkey: spotMarket['_decoded'].quoteVault },
                { isSigner: false, isWritable: true, pubkey: mangoGroup.vaults[i] },
                { isSigner: false, isWritable: true, pubkey: mangoGroup.vaults[mangoGroup.vaults.length - 1] },
                { isSigner: false, isWritable: false, pubkey: dexSigner },
                { isSigner: false, isWritable: false, pubkey: spl_token_1.TOKEN_PROGRAM_ID },
            ];
            const data = layout_1.encodeMangoInstruction({ SettleFunds: {} });
            const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
            transaction.add(instruction);
        }
        const deposits = marginAccount.getDeposits(mangoGroup);
        const liabs = marginAccount.getLiabs(mangoGroup);
        for (let i = 0; i < layout_1.NUM_TOKENS; i++) { // TODO test this. maybe it hits transaction size limit
            const deposit = deposits[i] + assetGains[i];
            if (deposit === 0 || liabs[i] === 0) {
                continue;
            }
            const keys = [
                { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
                { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
                { isSigner: true, isWritable: false, pubkey: owner.publicKey },
                { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY }
            ];
            const data = layout_1.encodeMangoInstruction({ SettleBorrow: { tokenIndex: new bn_js_1.default(i), quantity: utils_1.uiToNative(liabs[i] * 2, mangoGroup.mintDecimals[i]) } });
            const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
            transaction.add(instruction);
        }
        const additionalSigners = [];
        if (transaction.instructions.length == 0) {
            return null;
        }
        else {
            return await this.sendTransaction(connection, transaction, owner, additionalSigners);
        }
    }
    async liquidate(connection, programId, mangoGroup, liqeeMarginAccount, // liquidatee marginAccount
    liqor, // liquidator
    tokenAccs, depositQuantities) {
        const depositsBN = [];
        for (let i = 0; i < mangoGroup.tokens.length; i++) {
            depositsBN[i] = utils_1.uiToNative(depositQuantities[i], mangoGroup.mintDecimals[i]);
        }
        const keys = [
            { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
            { isSigner: true, isWritable: false, pubkey: liqor.publicKey },
            { isSigner: false, isWritable: true, pubkey: liqeeMarginAccount.publicKey },
            { isSigner: false, isWritable: false, pubkey: spl_token_1.TOKEN_PROGRAM_ID },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY },
            ...liqeeMarginAccount.openOrders.map((pubkey) => ({ isSigner: false, isWritable: false, pubkey })),
            ...mangoGroup.oracles.map((pubkey) => ({ isSigner: false, isWritable: false, pubkey })),
            ...mangoGroup.vaults.map((pubkey) => ({ isSigner: false, isWritable: true, pubkey })),
            ...tokenAccs.map((pubkey) => ({ isSigner: false, isWritable: true, pubkey })),
        ];
        const data = layout_1.encodeMangoInstruction({ Liquidate: { depositQuantities: depositsBN } });
        const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        const additionalSigners = [];
        return await this.sendTransaction(connection, transaction, liqor, additionalSigners);
    }
    async depositSrm(connection, programId, mangoGroup, marginAccount, owner, srmAccount, quantity) {
        const nativeQuantity = utils_1.uiToNative(quantity, token_instructions_1.SRM_DECIMALS);
        const keys = [
            { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
            { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
            { isSigner: true, isWritable: false, pubkey: owner.publicKey },
            { isSigner: false, isWritable: true, pubkey: srmAccount },
            { isSigner: false, isWritable: true, pubkey: mangoGroup.srmVault },
            { isSigner: false, isWritable: false, pubkey: spl_token_1.TOKEN_PROGRAM_ID },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY }
        ];
        const data = layout_1.encodeMangoInstruction({ DepositSrm: { quantity: nativeQuantity } });
        const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        const additionalSigners = [];
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    async withdrawSrm(connection, programId, mangoGroup, marginAccount, owner, srmAccount, quantity) {
        const nativeQuantity = utils_1.uiToNative(quantity, token_instructions_1.SRM_DECIMALS);
        const keys = [
            { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
            { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
            { isSigner: true, isWritable: false, pubkey: owner.publicKey },
            { isSigner: false, isWritable: true, pubkey: srmAccount },
            { isSigner: false, isWritable: true, pubkey: mangoGroup.srmVault },
            { isSigner: false, isWritable: false, pubkey: mangoGroup.signerKey },
            { isSigner: false, isWritable: false, pubkey: spl_token_1.TOKEN_PROGRAM_ID },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY }
        ];
        const data = layout_1.encodeMangoInstruction({ WithdrawSrm: { quantity: nativeQuantity } });
        const instruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        const additionalSigners = [];
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    async placeOrder(connection, programId, mangoGroup, marginAccount, spotMarket, owner, side, price, size, orderType, clientId) {
        // TODO allow wrapped SOL wallets
        orderType = orderType || 'limit';
        const limitPrice = spotMarket.priceNumberToLots(price);
        const maxBaseQuantity = spotMarket.baseSizeNumberToLots(size);
        const feeTier = serum_1.getFeeTier(0, utils_1.nativeToUi(mangoGroup.nativeSrm || 0, token_instructions_1.SRM_DECIMALS));
        const rates = serum_1.getFeeRates(feeTier);
        const maxQuoteQuantity = new bn_js_1.default(spotMarket['_decoded'].quoteLotSize.toNumber() * rates.taker).mul(spotMarket.baseSizeNumberToLots(size).mul(spotMarket.priceNumberToLots(price)));
        if (maxBaseQuantity.lte(new bn_js_1.default(0))) {
            throw new Error('size too small');
        }
        if (limitPrice.lte(new bn_js_1.default(0))) {
            throw new Error('invalid price');
        }
        const selfTradeBehavior = 'decrementTake';
        const marketIndex = mangoGroup.getMarketIndex(spotMarket);
        const vaultIndex = (side === 'buy') ? mangoGroup.vaults.length - 1 : marketIndex;
        // Add all instructions to one atomic transaction
        const transaction = new web3_js_1.Transaction();
        // Specify signers in addition to the wallet
        const additionalSigners = [];
        // Create a Solana account for the open orders account if it's missing
        const openOrdersKeys = [];
        for (let i = 0; i < marginAccount.openOrders.length; i++) {
            if (i === marketIndex && marginAccount.openOrders[marketIndex].equals(utils_1.zeroKey)) {
                // open orders missing for this market; create a new one now
                const openOrdersSpace = serum_1.OpenOrders.getLayout(mangoGroup.dexProgramId).span;
                const openOrdersLamports = await connection.getMinimumBalanceForRentExemption(openOrdersSpace, 'singleGossip');
                const accInstr = await utils_1.createAccountInstruction(connection, owner.publicKey, openOrdersSpace, mangoGroup.dexProgramId, openOrdersLamports);
                transaction.add(accInstr.instruction);
                additionalSigners.push(accInstr.account);
                openOrdersKeys.push(accInstr.account.publicKey);
            }
            else {
                openOrdersKeys.push(marginAccount.openOrders[i]);
            }
        }
        const keys = [
            { isSigner: false, isWritable: true, pubkey: mangoGroup.publicKey },
            { isSigner: true, isWritable: false, pubkey: owner.publicKey },
            { isSigner: false, isWritable: true, pubkey: marginAccount.publicKey },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_CLOCK_PUBKEY },
            { isSigner: false, isWritable: false, pubkey: spotMarket.programId },
            { isSigner: false, isWritable: true, pubkey: spotMarket.publicKey },
            { isSigner: false, isWritable: true, pubkey: spotMarket['_decoded'].requestQueue },
            { isSigner: false, isWritable: true, pubkey: spotMarket['_decoded'].eventQueue },
            { isSigner: false, isWritable: true, pubkey: spotMarket['_decoded'].bids },
            { isSigner: false, isWritable: true, pubkey: spotMarket['_decoded'].asks },
            { isSigner: false, isWritable: true, pubkey: mangoGroup.vaults[vaultIndex] },
            { isSigner: false, isWritable: false, pubkey: mangoGroup.signerKey },
            { isSigner: false, isWritable: true, pubkey: spotMarket['_decoded'].baseVault },
            { isSigner: false, isWritable: true, pubkey: spotMarket['_decoded'].quoteVault },
            { isSigner: false, isWritable: false, pubkey: spl_token_1.TOKEN_PROGRAM_ID },
            { isSigner: false, isWritable: false, pubkey: web3_js_1.SYSVAR_RENT_PUBKEY },
            { isSigner: false, isWritable: true, pubkey: mangoGroup.srmVault },
            ...openOrdersKeys.map((pubkey) => ({ isSigner: false, isWritable: true, pubkey })),
            ...mangoGroup.oracles.map((pubkey) => ({ isSigner: false, isWritable: false, pubkey })),
        ];
        const data = layout_1.encodeMangoInstruction({
            PlaceOrder: clientId
                ? { side, limitPrice, maxBaseQuantity, maxQuoteQuantity, selfTradeBehavior, orderType, clientId, limit: 65535 }
                : { side, limitPrice, maxBaseQuantity, maxQuoteQuantity, selfTradeBehavior, orderType, limit: 65535 }
        });
        const placeOrderInstruction = new web3_js_1.TransactionInstruction({ keys, data, programId });
        transaction.add(placeOrderInstruction);
        // sign, send and confirm transaction
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    async settleFunds(connection, programId, mangoGroup, marginAccount, owner, spotMarket) {
        const marketIndex = mangoGroup.getMarketIndex(spotMarket);
        const dexSigner = await web3_js_1.PublicKey.createProgramAddress([
            spotMarket.publicKey.toBuffer(),
            spotMarket['_decoded'].vaultSignerNonce.toArrayLike(Buffer, 'le', 8)
        ], spotMarket.programId);
        const instruction = instruction_1.makeSettleFundsInstruction(programId, mangoGroup.publicKey, owner.publicKey, marginAccount.publicKey, spotMarket.programId, spotMarket.publicKey, marginAccount.openOrders[marketIndex], mangoGroup.signerKey, spotMarket['_decoded'].baseVault, spotMarket['_decoded'].quoteVault, mangoGroup.vaults[marketIndex], mangoGroup.vaults[mangoGroup.vaults.length - 1], dexSigner);
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        // Specify signers in addition to the owner account
        const additionalSigners = [];
        // sign, send and confirm transaction
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    async cancelOrder(connection, programId, mangoGroup, marginAccount, owner, spotMarket, order) {
        const instruction = instruction_1.makeCancelOrderInstruction(programId, mangoGroup.publicKey, owner.publicKey, marginAccount.publicKey, spotMarket.programId, spotMarket.publicKey, spotMarket['_decoded'].bids, spotMarket['_decoded'].asks, order.openOrdersAddress, mangoGroup.signerKey, spotMarket['_decoded'].eventQueue, order);
        const transaction = new web3_js_1.Transaction();
        transaction.add(instruction);
        const additionalSigners = [];
        return await this.sendTransaction(connection, transaction, owner, additionalSigners);
    }
    async getMangoGroup(connection, mangoGroupPk, srmVaultPk) {
        if (srmVaultPk) {
            const [acc, srmVaultAcc] = await Promise.all([connection.getAccountInfo(mangoGroupPk), connection.getAccountInfo(srmVaultPk)]);
            const decoded = layout_1.MangoGroupLayout.decode(acc == null ? undefined : acc.data);
            if (!srmVaultAcc) {
                return new MangoGroup(mangoGroupPk, decoded);
            }
            const srmVault = utils_1.parseTokenAccountData(srmVaultAcc.data);
            return new MangoGroup(mangoGroupPk, decoded, srmVault.amount);
        }
        else {
            const acc = await connection.getAccountInfo(mangoGroupPk);
            const decoded = layout_1.MangoGroupLayout.decode(acc == null ? undefined : acc.data);
            return new MangoGroup(mangoGroupPk, decoded);
        }
    }
    async getMarginAccount(connection, marginAccountPk, dexProgramId) {
        const acc = await connection.getAccountInfo(marginAccountPk, 'singleGossip');
        const ma = new MarginAccount(marginAccountPk, layout_1.MarginAccountLayout.decode(acc == null ? undefined : acc.data));
        await ma.loadOpenOrders(connection, dexProgramId);
        return ma;
    }
    async getAllMarginAccounts(connection, programId, mangoGroup) {
        const filters = [
            {
                memcmp: {
                    offset: layout_1.MarginAccountLayout.offsetOf('mangoGroup'),
                    bytes: mangoGroup.publicKey.toBase58(),
                }
            },
            {
                dataSize: layout_1.MarginAccountLayout.span,
            },
        ];
        const marginAccountsProms = utils_1.getFilteredProgramAccounts(connection, programId, filters)
            .then((accounts) => (accounts.map(({ publicKey, accountInfo }) => new MarginAccount(publicKey, layout_1.MarginAccountLayout.decode(accountInfo == null ? undefined : accountInfo.data)))));
        const ordersFilters = [
            {
                memcmp: {
                    offset: serum_1.OpenOrders.getLayout(mangoGroup.dexProgramId).offsetOf('owner'),
                    bytes: mangoGroup.signerKey.toBase58()
                }
            },
            {
                dataSize: serum_1.OpenOrders.getLayout(mangoGroup.dexProgramId).span
            }
        ];
        const openOrdersProms = utils_1.getFilteredProgramAccounts(connection, mangoGroup.dexProgramId, ordersFilters)
            .then((accounts) => (accounts.map(({ publicKey, accountInfo }) => serum_1.OpenOrders.fromAccountInfo(publicKey, accountInfo, mangoGroup.dexProgramId))));
        const marginAccounts = await marginAccountsProms;
        const openOrders = await openOrdersProms;
        const pkToOpenOrdersAccount = {};
        openOrders.forEach((openOrdersAccount) => (pkToOpenOrdersAccount[openOrdersAccount.publicKey.toBase58()] = openOrdersAccount));
        for (const ma of marginAccounts) {
            for (let i = 0; i < ma.openOrders.length; i++) {
                if (ma.openOrders[i].toBase58() in pkToOpenOrdersAccount) {
                    ma.openOrdersAccounts[i] = pkToOpenOrdersAccount[ma.openOrders[i].toBase58()];
                }
            }
        }
        return marginAccounts;
    }
    async getMarginAccountsForOwner(connection, programId, mangoGroup, owner) {
        const filters = [
            {
                memcmp: {
                    offset: layout_1.MarginAccountLayout.offsetOf('mangoGroup'),
                    bytes: mangoGroup.publicKey.toBase58(),
                },
            },
            {
                memcmp: {
                    offset: layout_1.MarginAccountLayout.offsetOf('owner'),
                    bytes: owner.publicKey.toBase58(),
                }
            },
            {
                dataSize: layout_1.MarginAccountLayout.span,
            },
        ];
        const accounts = await utils_1.getFilteredProgramAccounts(connection, programId, filters);
        const marginAccounts = accounts.map(({ publicKey, accountInfo }) => new MarginAccount(publicKey, layout_1.MarginAccountLayout.decode(accountInfo == null ? undefined : accountInfo.data)));
        await Promise.all(marginAccounts.map((ma) => ma.loadOpenOrders(connection, mangoGroup.dexProgramId)));
        return marginAccounts;
    }
}
exports.MangoClient = MangoClient;
//# sourceMappingURL=client.js.map